#Update Let's Encrypt Certificates in IIS und ioBroker ber simpleAPI oder Command Line
#30.12.2017 by Daniel Matschull
#Version 1.3

################################################################
#                           Settings                           #
################################################################

#Let's Encrypt Parameter
$hostname = "mydomain.de"

#ioBroker Parameter
$ioBrokerPath = "C:\Program Files\ioBroker"				#ioBroker-Pfad
$ioBrokerServiceName = "ioBroker.exe"					#Name des ioBroker-Dienstes
$DPKey = "control.0.Certificates.UpdateKey"				#Datenpunkt fr Key-Update
$DPCert = "control.0.Certificates.UpdateCertificate"	#Datenpunkt fr Cert-Update
$SAPIhost = "https://iobrokerhost:6020/"				#Hostadresse und Port der ioBroker simpleAPI

#Update-Methode
$UpdMethod = "SAPI" 									#SAPI = simpleAPI oder CMD = ioBroker Commandline

#Zertifikate
$PEMCrtFile = "E:\Zertifikate\host.crt.pem"
$PEMKeyFile = "E:\Zertifikate\host.key.pem"
$CERFile = "E:\Zertifikate\host.cer"
$PFXFile = "E:\Zertifikate\host.pfx"

#IIS Parameter
$appid = "{xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx}"
$SiteName = "Webserver"
$FTPSiteName = "FTP-Server"

#Debug/Logging/Mail
$WriteLog = $true
$LogFilename = "D:\Logs\LECertUpdate.txt"
$MailRecipient = "webmaster@mydomain.de"
$MailSender = "Let's Encrypt Update Script <info@mydomain.de>"
$PSEmailServer = "localhost"

#Powershell-Parameter
$ErrorActionPreference = "Stop"

################################################################
########################## Funktionen ##########################
################################################################

function sendCommand($method, $action, $dp, $value) {

    #Befehl zusammen bauen
    $command = $action + "/" + $dp

	#Wert angegeben und Methode GET, Wert in URL als Parameter anhngen
	if ($value -and $method -eq "GET") {
		$command = $command + "?value=" + $value
    }

    Log ("Sende Request: " +  $SAPIhost + $command)

    $url = $SAPIhost + $command;

    try {
        #bei POST Daten mitsenden, sonst nicht
        if ($method -eq "POST") {
            $response = Invoke-WebRequest $url -Method Post -Body $value
        }
        else {
            $response = Invoke-WebRequest $url
        }

        Log ("`tHTTP-Status: " + $response.StatusCode)
        Log ("`tContent: " + $response.Content)

        #Request erfolgreich?
        if ($response.StatusCode -eq 200) {
            return $response.Content
        }
        else {
            return $false
        }
    }
    catch {
        #Response nicht leer?
        if ($_.Exception.Response) {
            Log ("`tHTTP-Status: " + $_.Exception.Response.StatusCode.value__)
            Log ("`tDescription: " + $_.Exception.Message)
            
        }
        else {
            Log ("`tStatus: " + $_.Exception.Status)
            Log ("`tDescription: " + $_.Exception.Message)
        }
        return $false
    }
}

#Update eines Datenpunkts per SAPI oder CMD
function UpdateDP($dp, $value) {

    #per simpleAPI
    if ($UpdMethod -eq "SAPI") {
        
        $result = sendCommand "POST" "setValueFromBody" $dp $value
        if ($result) {
            $success = $true;
        }
        else {
            $success = $false;
        }
    }
    else {
    #per Command Line
        #Command zusammen bauen
        $args = ("state", "set", $dp, ("`" " + $value + "`""))
        $runFile = ".\ioBroker.bat"

        #Arbeitsverzeichnis wechseln
        cd $ioBrokerPath

        Log ("Befehl ausfhren: " + $args)

        #Befehl ausfhren
        try {
            $result = & $runFile $args 2>&1
        }
        catch {
            Log ("ERROR: "+ $_Exception.Message)
        }

        #auf Fehler prfen
        if ($result.Exception) {
            $success = $false
        }
        else {
            if ($result -like "Set `"" + $dp + "`" with value:*") {
                $success = $true
            }
            else {
                $success = $false
            }
        }
    }
    #frs Logging
	if ($success) {
		Log ("Update `"" +  $dp + "`" erfolgreich!")
    }
	else {
		Log ("Fehler beim Updating von `"" + $dp + "`"!")
    }
    
    return $success
}

#PEM-Zertifikat einlesen
function readPEMFile($file) {

    Log ("Lese PEM-File: " + $file)

    try {
        $data = Get-Content $file -ErrorAction Stop

        #gelesene Zeilen mit "\n" in String umwandeln (fr druckbaren Zeilenumbruch)
        return $data -join "\n"
    }
    catch {
        Log ("ERROR: " + $_.Exception.Message)
        #Skript wird beendet
        exit -1
    }
}

#prfen, ob ioBroker bereits luft
function checkioBroker {

    #Dienste ermiteln
    $foundService = (Get-Service | Where-Object Name -match $ioBrokerServiceName)

    #existiert der Dienst?
    if ($foundService)
    {    
        return $foundService.Status
    }
    else {
        Log ("ERROR: Der Dienst `"" + $ioBrokerServiceName + "`" wurde nicht gefunden!")
        return $false
    }
}

#Startet ioBroker-Dienst neu bzw. startet ihn, wenn ioBroker nicht luft
function restartioBroker($action) {

    try {
        #nur starten
        if ($action -eq "start") {
            #luft ioBroker schon?
            if ((checkioBroker) -eq "Running") {
                Log "ioBroker luft bereits!"
            }
            else {
                Log "ioBroker wird gestartet..."

                Start-Service $ioBrokerServiceName

                Log "ioBroker wurde gestartet!"
            }
        }
        else {
        #sonst neu starten
            Log "ioBroker wird neu gestartet..."

            Restart-Service $ioBrokerServiceName

            Log "ioBroker wurde neu gestartet!"
        }
    }
    catch {
        Log ("ERROR: " + $_.Exception.Message)
    }
}


#Schreibt einen Logeintrag
function log($text) {

    $loglen = 150

    if ($WriteLog)
    {
        if ($text.Length -gt $loglen)
        {
            $logtext =  $text.Substring(0, $loglen) + "..."
        }
        else
        {
            $logtext = $text
        }
        Add-Content $LogFilename -value ((Get-Date -Format "dd.MM.yyyy HH:mm:ss: ") + $logtext)
    }
}

#Sendet Mails
function SendMail($text)
{
    Send-MailMessage -to $MailRecipient -from $MailSender -Subject "Update der LE-Zertifikate" -body $text
}

################################################################
######################## Hauptfunktionen #######################
################################################################

#neues LE-Zertifikat anfordern
function renewLECertificate()
{
    # ACMESharp-Modul laden
    Import-Module ACMESharp

    try
    {
        #Aliasname und CertName erzeugen
        $DNSalias = "DNS" + $hostname + "_" + (Get-Date -Format yyyy-MM-dd)
        $CertAlias = "CERT" + $hostname + "_" + (Get-Date -Format yyyy-MM-dd)

        #neuen Identifier erzeugen
        Log (">Neuen DNS-Alias `"" + $DNSalias + "`" fr DNS `"" + $hostname + "`" wird erzeugt...")
        New-ACMEIdentifier -Dns $hostname -Alias $DNSalias
        Log "OK"

        #Prfen ob Identifier bereits gltig ist (z.B. noch nicht abgelaufen)
        if ((Update-ACMEIdentifier $DNSalias | Select-Object Status).Status -eq("valid"))
        {
            Log ("DNS-Alias `"" + $DNSalias + "`" fr Hostname `"" + $hostname + "`" ist bereits gltig. Fahre fort...")
        }
        else
        {
            #Challenge Handling
            Log (">Challenge Handling fr DNS-Alias `"" + $DNSalias + "`" wird fr IIS-Site `"" + $SiteName + "`" wird eingerichtet...")
            Complete-ACMEChallenge $DNSalias -ChallengeType http-01 -Handler iis -HandlerParameters @{ WebSiteRef = "$SiteName" }
            Log "OK"

            #Challenge durchfhren
            Log (">Challenge fr DNS-Alias `"" + $DNSalias + "`" wird durchgefhrt...")
            Submit-ACMEChallenge $DNSalias -ChallengeType http-01
            Log "OK. Warte 30 Sekunden..."

            #30 Sekunden warten bis Challenge durchgefhrt
            Start-Sleep 30
            #Update Challenge/Status prfen
            if (-Not ((Update-ACMEIdentifier $DNSalias | Select-Object Status).Status -eq("valid")))
            {
                Log ("ERROR: Challenge fr DNS-Alias `"" + $DNSalias + "`" fehlgeschlagen!")
                return $false
            }
            else
            {
                Log ("Challenge fr DNS-Alias `"" + $DNSalias + "`" erfolgreich!")
            }
        }

        #neues Zertifikat anfordern
        Log (">Neues Zertikat mit dem Alias `"" + $CertAlias + "`" anfordern...")
        New-ACMECertificate $DNSalias -Generate -Alias $CertAlias
        Log "OK"

        #Anforderung absenden
        Log (">Zertikatsanforderung mit dem Alias `"" + $CertAlias + "`" wird abgeschickt...")
        Submit-ACMECertificate $CertAlias
        Log "OK. Warte 60 Sekunden..."

        #1 Minute warten bis Zertifikat ausgestellt wurde
        Start-Sleep 60
        #Zertifiktsstatus aktualisieren/Status prfen
        if ((Update-ACMECertificate $CertAlias | Select-Object SerialNumber).SerialNumber -eq $null)
        {
            Log ("ERROR: Ausstellen des Zertifikats mit dem Alias `"" + $CertAlias + "`" fehlgeschlagen!")
            return $false
        }
        else
        {
            Log ("Ausstellen des Zertifikats mit dem Alias `"" + $CertAlias + "`" erfolgreich!")
        }

        #Export Private Key (PEM)
        Log (">PEM-Key fr Alias `"" + $CertAlias + "`" wird nach `"" + $PEMKeyFile + "`" exportiert...")
        Get-ACMECertificate $CertAlias -ExportKeyPEM $PEMKeyFile -overwrite
        Log "OK"

        #Export Certificate (PEM+CER)
        Log (">PEM-Zertifikat fr Alias `"" + $CertAlias + "`" wird nach `"" + $PEMCrtFile + "`" exportiert.")
        Get-ACMECertificate $CertAlias -ExportCertificatePEM $PEMCrtFile -ExportCertificateDER $CERFile -overwrite
        Log "OK"

        #Export PKCS#12 (PFX) Archive
        Log (">PFX-Zertifikat fr Alias `"" + $CertAlias + "`" wird nach `"" + $PFXFile + "`" exportiert.")
        Get-ACMECertificate $CertAlias -ExportPkcs12 $PFXFile -CertificatePassword "zion" -overwrite
        Log "OK"

		return $true
    }
    catch
    {
        Log ("ERROR: " + $_.Exception.Message)
    }
    return $false	#Skript mit Fehler beendet
}

#IIS-Bindungen aktualisieren
function UpdateIISCertificates()
{
    #Altes Zertifkat entfernen
    Log (">Altes PFX-Zertifikat zum Host `"" + $hostname + "`" wird gelscht...")
    $result = certutil -delstore my $hostname
    Log $result

    #Neues PFX-Zertifikat importieren
    Log (">Neues PFX-Zertifikat `"" + $PFXFile + "`" wird importiert...")
    $result = certutil -f -p "zion" -importpfx $PFXFile
    Log $result

    #Hash das Importierten Zertifikats holen
    Log ">Hash des importierten Zertifikats ermitteln..."
    $certhash = (Get-ChildItem cert:\LocalMachine\My | Where-Object Subject -like "*$hostname"  | Select-Object -First 1).Thumbprint
    Log ("Hash=" + $certhash)

    #aktuelle SSL-Zertifikats-Bindung lchen
    Log ">SSL-Zertifikatsbindungen in IIS lschen..."
    Log "Fr Bindung 443..."
    $result = netsh http delete sslcert ipport=0.0.0.0:443
    Log $result
    Log "Fr Bindung 4443..."
    $result = netsh http delete sslcert ipport=0.0.0.0:4443
    Log $result

    #neues Zertifikat binden
    Log ">Neues Zertifikat in IIS binden..."
    Log "Fr Bindung 443..."
    $result = netsh http add sslcert ipport=0.0.0.0:443 certhash=$certhash appid=$appid
    Log $result
    Log "Fr Bindung 4443..."
    $result = netsh http add sslcert ipport=0.0.0.0:4443 certhash=$certhash appid=$appid
    Log $result

    #neues Zertifikat fr FTPS binden
    Log ">Neues Zertifikat in IIS fr FTPS binden..."
    $filter = "/system.applicationHost/sites/site[@Name='" + $FTPSiteName + "']/ftpserver/security/ssl"
    Set-WebConfigurationProperty -filter $filter -name serverCertHash -Value $certhash -WarningVariable warn
    if ($warn.Count -gt 0) { Log $warn }
}

#ioBroker-Zertifikat und Key aktualisieren
function UpdateioBrokerCertificates()
{
    $err = $false
    #prfe, ob ioBroker gestartet ist
    if ((checkioBroker) -eq "Running") {
        Log "ioBroker luft!"
    }
    else {
        Log "ioBroker luft nicht!"
        restartioBroker("start")
    }

    #Zertifikate in Sting lesen
    $PEMKeyString = readPEMFile($PEMKeyFile)
    $PEMCrtString = readPEMFile($PEMCrtFile)

    #Update der Zertifikate
    #Cert
    Log ">Updating Cert..."
    if (-Not (UpdateDP $DPCert $PEMCrtString)) { $err = $true }
    #Key
    Log ">Updating Key..."
    if (-Not (UpdateDP $DPKey $PEMKeyString)) { $err = $true }

    #wenn kein Fehler, dann ioBroker neun starten
    if (-Not $err) {
        Start-Sleep 5  #5 Sekunden warten
        restartioBroker
    }
}

################################################################
############################# Main #############################
################################################################

#Beginn Log
Log "-----Let's Encrypt CertUpdate Script gestartet-----"

##################### Let's Encrypt Stuff ######################
Log ">>>Neues Let's Encrypt Zertifikat anfordern<<<"
$LEresult = renewLECertificate
if (-Not ($LEresult[$LEresult.count-1]))
{
    SendMail "Fehler beim Anfordern des neuen LE-Zertifikats!`nUpdate-Vorgang wird abgebrochen!`nSiehe Log-Datei!"
    Log "Skript wird abgebrochen!"
    exit -1
}

#################### Update IIS Zertifikate ####################
Log ">>> IIS Bindungen aktualisieren<<<"
UpdateIISCertificates

################## Update ioBroker Zertifikate #################
Log ">>> ioBroker Zertifikate aktualisieren<<<"
UpdateioBrokerCertificates

#Ende Log
Log "-----Let's Encrypt CertUpdate Script beendet-----"
SendMail "Update der LE-Zertifikate erfolgreich!"
exit 0